'''
Copyright (C) 2019 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import math

from bl_math import clamp, lerp
from mathutils import Vector, kdtree, Matrix, Quaternion
# from mathutils.bvhtree import BVHTree
from gpu_extras.batch import batch_for_shader
from bpy_extras import view3d_utils
import gpu
import bmesh
import blf
from os import path
from mathutils.geometry import intersect_point_line, distance_point_to_plane
import traceback
from bpy.app.handlers import persistent

from . import draw_xray # for global TARGET_BVH_LIST - holds BVH for snap objects
from .draw_xray import get_obj_mesh_bvht

def log(*args, **kwargs):
    return
    print(*args, **kwargs)

def close_bmesh( context, bm, source):
    bm.normal_update()
    if context.object.mode == "EDIT":
        bmesh.update_edit_mesh(source, loop_triangles=False, destructive=False)
    else:
        bm.to_mesh(source)
        bm.free()

def get_mirrored_verts(mesh):
    return # not used yet..
    kd = kdtree.KDTree(len(mesh.vertices))
    for i, v in enumerate(mesh.vertices):
        kd.insert(v.co, i)
    kd.balance()

    mirror_map = []
    for vert in mesh.vertices:
        co, mirror_vert_index, dist = kd.find(Vector((-vert.co[0], vert.co[1], vert.co[2])))
        if mirror_vert_index == vert.index:
            mirror_map.append(-1)
        elif dist < 0.0001:  # delta
            mirror_map.append(mirror_vert_index)
        else:
            mirror_map.append(-2)
    return tuple(mirror_map)


# round_to_n = lambda x, n: round(x, -int(math.floor(math.log10(x))) + (n - 1))
def round_to_first_digit(x):
    return math.pow(10, math.floor(math.log10(x)))


def raycast(context, mouse_2d_co):
    if not draw_xray.TARGET_BVH_LIST:
        depsgraph = context.evaluated_depsgraph_get()
        draw_xray.TARGET_BVH_LIST = get_obj_mesh_bvht(context, depsgraph, applyModifiers=True, world_space=True) # in reality in active obj space...

    if not draw_xray.TARGET_BVH_LIST: # no target was set, or target is not visible..draw_xray.TARGET_BVH_LIST
        return None, None

    region = context.region
    rv3d = context.region_data
    ray_max = 1000.0
    mat_inv = context.active_object.matrix_world.inverted()

    # get the ray from the viewport and mouse
    view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, mouse_2d_co)
    ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, mouse_2d_co)

    if rv3d.view_perspective == 'ORTHO':  # move ortho origin back
        ray_origin = ray_origin - (view_vector * (ray_max / 2.0))

    ray_target = ray_origin + (view_vector * ray_max)

    ray_origin_obj = mat_inv @ ray_origin  # not sure why this order works, whatever...
    ray_target_obj = mat_inv @ ray_target
    view_vector_obj = ray_target_obj - ray_origin_obj

    # ray_direction = ray_origin - ray_target
    hit_local, normal, face_index, depth = draw_xray.TARGET_BVH_LIST.ray_cast(ray_origin_obj, view_vector_obj)
    return hit_local, normal

def detect_hole_adjacent_to_edge(edge):
    max_size = 4 # check max for 4 edges
    hole_edges = [edge]
    step = 0
    current_edge = edge
    vert = current_edge.verts[0]
    # how about if loop was made but around of face?
    if len(current_edge.link_faces) == 0:
        return
    e_face = current_edge.link_faces[0] # loop around - e.faces would not be same
    while True:
        next_edge = [e for e in vert.link_edges if e != current_edge and e.is_boundary and e.link_faces[0] != e_face]
        if not next_edge:
            return
        if len(next_edge) > 1:
            return
        current_edge = next_edge[0]
        e_face = current_edge.link_faces[0]
        if current_edge == edge: # made loop
            return hole_edges
        hole_edges.append(current_edge)
        vert = current_edge.other_vert(vert)
        step += 1
        if step > max_size:
            return None

    return hole_edges

def get_view_vec(context):
    rv3d = context.region_data
    view_mat = rv3d.view_matrix.to_3x3()
    # screen_depth = view_ma[2].xyz
    # screen_up = view_ma[1].xyz
    # screen_right = view_ma[1].xyz
    return view_mat[2].xyz

def circle_coords_calc(circle_resol):
    m_2pi = 2 * math.pi
    coords = ()
    for a in range(circle_resol + 1):
        ang = (m_2pi * a) / circle_resol
        coords += (math.sin(ang), math.cos(ang), 0.0),
    return coords

circle_co = circle_coords_calc(32)

shader_3d_uniform = gpu.shader.from_builtin('UNIFORM_COLOR')


def draw_circle_px(self, color):
    gpu.state.line_width_set(2)
    gpu.state.blend_set("ALPHA")
    gpu.state.depth_test_set("NONE")
    with gpu.matrix.push_pop():
        gpu.matrix.multiply_matrix(self.matrix_basis)
        batch = batch_for_shader(shader_3d_uniform, 'LINE_LOOP', {"pos": circle_co})
        shader_3d_uniform.bind()
        shader_3d_uniform.uniform_float("color", color)
        batch.draw(shader_3d_uniform)

    gpu.state.blend_set("NONE")
    gpu.state.line_width_set(1)


SIDE_PTS = []
TRACK_PTS = [] # two track points for edges
HOVER_EDGE = None
# function that draws gls text in 2d mouse position
def draw_points(positions, color):
    batch = batch_for_shader(shader_3d_uniform, 'POINTS', {"pos": positions})
    shader_3d_uniform.bind()
    shader_3d_uniform.uniform_float("color", color)
    batch.draw(shader_3d_uniform)

def draw_lines(positions, color):
    batch = batch_for_shader(shader_3d_uniform, 'LINES', {"pos": positions})
    shader_3d_uniform.bind()
    shader_3d_uniform.uniform_float("color", color)
    batch.draw(shader_3d_uniform)


col1 = (0.7, 1, 0, 1)
col2 = (0, 1, 0.7, 1)
white = (1, 1, 1, 1)

def draw_debug_px(self, color):
    gpu.state.blend_set("ALPHA")
    gpu.state.line_width_set(3)
    gpu.state.point_size_set(5)

    with gpu.matrix.push_pop():
        gpu.matrix.multiply_matrix(bpy.context.active_object.matrix_world)

        if SIDE_PTS:
            draw_points([SIDE_PTS[0]], col1)
            draw_points([SIDE_PTS[1]], col2)

        if TRACK_PTS:
            if TRACK_PTS[0]:
                draw_points([TRACK_PTS[0]], col1)
            if TRACK_PTS[1]:
                draw_points([TRACK_PTS[1]], col2)

        if self.track_border_edge1:
            v1, v2 = self.track_border_edge1.verts
            draw_lines([v1.co, v2.co], col1)

        if self.track_border_edge2:
            v1, v2 = self.track_border_edge2.verts
            draw_lines([v1.co, v2.co], col2)

        if self.strip_perp_edges:
            v1, v2 = self.strip_perp_edges[-1].verts[0], self.strip_perp_edges[-1].verts[1]
            draw_lines([v1.co, v2.co], white)

        if HOVER_EDGE and HOVER_EDGE.is_valid:
            v1, v2 = HOVER_EDGE.verts
            draw_lines([v1.co, v2.co], white)

        if self.hit_local:
            draw_points([self.hit_local], white)

            # tool = bpy.context.workspace.tools.from_space_view3d_mode(bpy.context.mode)
            # tool_props = tool.operator_properties("mesh.quad_xray_brush")
            # tool.radius
            if self.tangent:  # line from prev_hit to hit_local to prev + TAN
                scale_factor = bpy.context.active_object.scale.length/math.sqrt(3)
                gpu.state.point_size_set(4)
                draw_lines([self.hit_local, self.hit_local + self.tangent * self.radius/scale_factor], (0, 1, 1, 1))
                draw_points([self.drag_co], (1, 0, 0, 1))

    gpu.state.blend_set("NONE")

def draw_text_px(self, color):
    gpu.state.blend_set("ALPHA")
    font_id = 0
    blf.color(font_id, *color)
    blf.position(font_id, self.mouse_init_co[0]-20, self.mouse_init_co[1], 0)
    blf.size(font_id, 30)
    # blf.draw(font_id, str(int(self.strength+0.5)))
    gpu.state.blend_set("NONE")


# print(path.join(path.dirname(__file__),"icons/draw_xray.draw_quads"))

class ToolXrayQuads(bpy.types.WorkSpaceTool):
    bl_space_type = 'VIEW_3D'
    bl_context_mode = 'EDIT_MESH'
    bl_idname="xray_tool.draw_quads"
    bl_label="Draw Xray Quads"
    bl_description="Draw Xray Quads"
    # bl_icon="ops.mesh.vertices_smooth"
    bl_icon = path.join(path.dirname(__file__),"icons/draw_xray.draw_quads")
    bl_widget="MESH_GGT_brush_circle_xray"

    bl_keymap = (
        ("mesh.quad_xray_brush", {"type": 'LEFTMOUSE', "value": 'PRESS'}, {}),
        # ("mesh.quad_xray_brush", {"type": 'LEFTMOUSE', "value": 'PRESS'}, {"properties": [("draw", True), ]}),
        ("mesh.quad_xray_brush", {"type": 'F', "value": 'PRESS', 'shift': True}, {"properties": [("resize", True), ]}),
        # ("mesh.quad_xray_brush", {"type": 'F', "value": 'PRESS', 'shift': True}, {"properties": [("strengthen", True), ]}),
    )

    def draw_settings(context, layout, tool):
        xray_brush_props = tool.operator_properties("mesh.quad_xray_brush")
        row = layout.row(align=True)
        row.prop(xray_brush_props, 'auto_radius', toggle=True)
        if not xray_brush_props.auto_radius:
            row.prop(xray_brush_props, 'radius')
        row.separator()
        # if not xray_brush_props.auto_spacing_factor:
        row.prop(xray_brush_props, 'spacing_factor', icon='FIXED_SIZE')
        row.prop(xray_brush_props, 'auto_spacing_factor', toggle=True)
        row.separator()
        row.prop(xray_brush_props, 'split_undo', toggle=True)
        # layout.prop(xray_brush_props, 'strength')


# NOTE: gizmo wont show up when LMB is clicked (for this we draw using modal drawing)
class XrayCircleCustomShapeWidget(bpy.types.Gizmo):
    bl_idname = "VIEW3D_GT_Circle_Gizmo_Xray"
    __slots__ = (
        "last_depth",
        "first_run",  # do not run gizmo draw till we inith matrix_basis
    )

    def update_matrix(self, context, mouse_co):
        # print("Mouse Co", mouse_co)
        global GIZMO_MOUSE_CO
        size = 3 # delete if GIZMO_MOUSE_CO has more than 4 itmes
        if len(GIZMO_MOUSE_CO) > size:
            GIZMO_MOUSE_CO.pop(0)
        GIZMO_MOUSE_CO.append(mouse_co)

        hit_local, normal = raycast(context, mouse_co)
        hover_over_mesh = hit_local is not None
        if hit_local is None:
            rv3d = context.region_data
            normal = get_view_vec(context)
            hit_local = view3d_utils.region_2d_to_location_3d(context.region, rv3d, mouse_co, self.last_depth)

        self.last_depth = hit_local.copy()
        quat = normal.to_track_quat('Z','Y')  # to_track_quat(track, up)
        rot_mat = quat.to_matrix().to_4x4()

        #NOTE get circle radius, from vps_tool radius prop
        tool = context.workspace.tools.from_space_view3d_mode(context.mode)
        tool_props = tool.operator_properties("mesh.quad_xray_brush")
        rad_scale_mat = Matrix.Scale(tool_props.radius, 4)

        if hover_over_mesh:
            hit_world = context.active_object.matrix_world @ hit_local
            # self.matrix_basis = context.active_object.matrix_world @ Matrix.Translation(hit_local) @ rot_mat @ rad_scale_mat
            self.matrix_basis = Matrix.Translation(hit_world) @ rot_mat @ rad_scale_mat
        else:
            self.matrix_basis = Matrix.Translation(hit_local) @ rot_mat @ rad_scale_mat

    def draw(self, context):
        # print("Gizmo draw")
        if not self.first_run: #avoids drawing brush circle at Vector((0, 0, 0))
            draw_circle_px(self, (0.9,0.9,0.9,1))

    # def draw_select(self, context, select_id): #XXX breaks test_select !
    #     mat = context.active_object.matrix_world
    #     self.draw_preset_circle(mat, axis='POS_Z', select_id=select_id)

    def setup(self):
        # print("Gizmo setup")
        self.first_run = True
        self.use_draw_offset_scale = True
        self.last_depth = Vector((0, 0, 0))

    # def invoke(self, context, event): #seems it is not executed eve onec..
    #     # print("Gizmo invoke !!")
    #     return {'RUNNING_MODAL'}

    # def exit(self, context, cancel): #seems it is not executed eve onec..
    #     print("gizmo Exit")

    def test_select(self, context, location):
        # makes it possible to hover highlight, wont work f draw_select() is used at same time
        self.update_matrix(context, location)
        self.first_run = False # gizmo matrix basis was set, now we can draw gizmo
        context.area.tag_redraw()
        return -1

    # def modal(self, context, event, tweak): #seems it is not executed eve onec..
    #     print('Gizmo modal')
    #     return {'RUNNING_MODAL'}

@persistent
def check_obj_updated(scene, depsgraph):
    if not bpy.context.active_object:
        return
    if bpy.context.mode != 'EDIT_MESH': # Only fire cache update - when jumping to edit mode.
        # print("Xray: clearing bvh")
        draw_xray.TARGET_BVH_LIST = None


def enable_bvh_check() -> None:
    if check_obj_updated not in bpy.app.handlers.depsgraph_update_post:
        bpy.app.handlers.depsgraph_update_post.append(check_obj_updated)


def disable_bvh_check():
    if check_obj_updated in bpy.app.handlers.depsgraph_update_post:
        bpy.app.handlers.depsgraph_update_post.remove(check_obj_updated)
    # draw_xray.TARGET_BVH_LIST = None

def is_quad_draw_on(context):
    return bpy.context.workspace.tools.from_space_view3d_mode(mode=context.mode).idname == ToolXrayQuads.bl_idname


GIZMO_MOUSE_CO = [] # so we can calc Tangent on init()
LOCK_GIZMO_DRAW = False # lock gizmo when loading new blend - or else crash

@persistent
def load_pre_xray_gizmo_fix(scene):
    '''  this prevet crash when using Xray draw strips  worksapce tool and loading new blend '''
    global LOCK_GIZMO_DRAW
    LOCK_GIZMO_DRAW = True

@persistent
def load_post_xray_gizmo_fix(scene):
    global LOCK_GIZMO_DRAW
    LOCK_GIZMO_DRAW = False


IS_HANDLER_ON = False # timed register are too slow, thus use global flag instead
UNLINK_DELAYED = False #  so  we run disable_bvh_check' on poll False

class XrayBrushCircleGizmoGroup(bpy.types.GizmoGroup):
    bl_idname = "MESH_GGT_brush_circle_xray"
    bl_label = "Side of Plane Gizmo"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'
    bl_options = {'3D'}


    def __init__(self):  # enabled once on worktool ON -- on some scenes it seems to fire 2x, cos __del__ is fired, after first __init__ ...
        if  LOCK_GIZMO_DRAW: # prevent  crash on new blend load - with Grid Tool enabled
            return
        # print('__init__')
        global IS_HANDLER_ON
        if not IS_HANDLER_ON and is_quad_draw_on(bpy.context):
            # print('enabling xray bvh handle')
            # NOTE: because of unlink delayed?
            IS_HANDLER_ON = True
            bpy.app.timers.register(enable_bvh_check, first_interval=0.2) # or crash when BloobTool disabled, and loading blend with it enabled

    def __del__(self):   # issue - blender spams __del__ multiple times. bunch of hacks to avoid it
        # enabled once on worktool OFF
        if LOCK_GIZMO_DRAW:   # prevent  crash on new blend load - with Grid Tool enabled
            return
        # print('__del__')
        workspace = bpy.context.workspace # gizmo disable check hack - so we do not run __dell__ on each draw stroke...
        global UNLINK_DELAYED, IS_HANDLER_ON
        if IS_HANDLER_ON and (workspace or UNLINK_DELAYED):
            UNLINK_DELAYED = False
            IS_HANDLER_ON = False
            # print('disabling xray bvh handle')
            bpy.app.timers.register(disable_bvh_check, first_interval=0.2) # or crash when BloobTool disabled, and loading blend with it enabled


    @classmethod
    def poll(cls, context):
        is_xray_draw_tool_on = context.mode == 'EDIT_MESH' and context.workspace.tools.from_space_view3d_mode(mode="EDIT_MESH").idname == "xray_tool.draw_quads"
        if not is_xray_draw_tool_on or not context.scene.xray_gizmo_show:
            global UNLINK_DELAYED
            UNLINK_DELAYED = context.scene.xray_gizmo_show # do not mark UNLINK_DELAYED if gizmo was temp disabled - for modal ops quad_draw

            wm = context.window_manager
            wm.gizmo_group_type_unlink_delayed(XrayBrushCircleGizmoGroup.bl_idname)
            return False
        return True

    def setup(self, context): #executed once on gizmo start
        self.widget = self.gizmos.new("VIEW3D_GT_Circle_Gizmo_Xray")

    # def draw_prepare(self, context): #we can update gizmo here, eg mat_basisa
    #     pass

    # def __del__(self):
    #     if hasattr(self, "widget"):
    #         pass
    #         # object.__getattribute__(self.widget, 'finish_gizmo')() # same as self.widget.finish_gizmo() ??


def laplacian_smooth(verts, factor=0.5):
    new_positions = []
    for vert in verts:
        if not vert.is_valid:
            continue
        neighbor_positions = [e.other_vert(vert).co for e in vert.link_edges if e.is_boundary]
        if not neighbor_positions:
            new_positions.append(vert.co)
            continue
        average_position = sum(neighbor_positions, Vector()) / len(neighbor_positions)
        new_position = vert.co.lerp(average_position, factor)
        new_positions.append(new_position)

    for vert, new_position in zip(verts, new_positions):
        if vert.is_valid:
            vert.co = new_position

def scale_vector(prev_edge_center, hit_local, radius):
    vector = hit_local - prev_edge_center
    scaling_factor = 2 * radius / vector.length
    scaled_vector = vector * scaling_factor
    return prev_edge_center + scaled_vector


def calc_edge_center(edge):
    return (edge.verts[0].co + edge.verts[1].co) / 2


identity_quat = Quaternion((1, 0, 0, 0))

class XRAY_OT_DrawQuads(bpy.types.Operator):
    bl_idname = "mesh.quad_xray_brush"
    bl_label = "Draw Quads Burus"
    bl_description = "Draw Quads Burus"
    bl_options = {"REGISTER", "UNDO"}

    resize: bpy.props.BoolProperty(name='resize', description='', default=False) # resize mode

    radius: bpy.props.FloatProperty(name="Radius", description="Brush Radius", default=0.5, soft_min=0.001, soft_max=1)
    spacing_factor: bpy.props.FloatProperty(name="Spacing", description="Segment spacing as fraction of radius size", default=1.0, soft_min=0.2, soft_max=2, subtype='FACTOR')
    auto_radius: bpy.props.BoolProperty(name="Auto Radius", description="Automatically match Radius size to edge length", default=True)
    auto_spacing_factor: bpy.props.BoolProperty(name="Auto Spacing", description="Automatically match segment spacing to quad length", default=True)
    split_undo: bpy.props.BoolProperty(name="Split Undo", description="Split undo steps for each drawn quad", default=False)

    # mirror_merge_dist:bpy.props.FloatProperty(name="Merge Distance", description="Distance to merge vertices", default=0.001, min=0.0, subtype='DISTANCE')

    _circel2d_handle = None
    _ui3d_handle = None

    def weighted_drag_smooth(self, active_obj, pos, rad):
        # simulate moving of self.drag_co - behind pos (usually behind mouse hit)
        to_drag = self.drag_co - pos
        length = to_drag.length

        # make sure we are in radius distance behind pos
        if length > rad:
            to_drag = to_drag/length * rad  if length != 0 else pos # same as to_hit.normalized() * dist
        self.drag_co = pos + to_drag


    def weld_strip_vertices(self, context, radius): # XXX:  not used
        #  XXX: not used
        weld_threshold = 0.6 * radius

        # Collect all vertices in the strip
        strip_verts = set()
        for edge in self.strip_perp_edges:
            strip_verts.update(edge.verts)

        # Create KD-tree for fast nearest neighbor search, excluding strip vertices
        kd = kdtree.KDTree(len(self.bm.verts) - len(strip_verts))
        for i, v in enumerate(self.bm.verts):
            if v not in strip_verts:
                kd.insert(v.co, i)
        kd.balance()

        # Find nearby vertices for each vertex in the strip
        verts_to_merge = {}
        for vert in strip_verts:
            co, index, dist = kd.find(vert.co)
            if dist < weld_threshold:
                target_vert = next(v for v in self.bm.verts if v.co == co and v not in strip_verts)
                verts_to_merge[vert] = target_vert

        if verts_to_merge:
            bmesh.ops.weld_verts(self.bm, targetmap=verts_to_merge)
            # Clean up degenerate edges and faces
            # bmesh.ops.remove_doubles(self.bm, verts=self.bm.verts, dist=0.0001)
            bmesh.update_edit_mesh(context.active_object.data)


    @staticmethod
    def update_edge_midpoints_kdtree(bm, max_dist):
        '''Build kd tree from sliced edges - without endings (eg. only mid point)
           Max distance is used to slice edges into smaller segments, and build kd tree from them'''
        boundary_edges = [e for e in bm.edges if e.is_boundary]

        kd_map = {}
        kd_cnt = 0
        for edge in boundary_edges:
            v0, v1 = edge.verts[0].co, edge.verts[1].co
            edge_length = (v1 - v0).length
            num_slices = max(1, int(edge_length / max_dist))  # Ensure at least one slice - int - floor
            kd_map[edge.index] = [v0.lerp(v1, (i + 1) / (num_slices + 1)) for i in range(num_slices)]
            kd_cnt += num_slices

        kd = kdtree.KDTree(kd_cnt)
        for ei, pos_list in kd_map.items():
            for pos in pos_list:
                kd.insert(pos, ei)
        kd.balance()
        return kd


    def find_closest_boundary_edge(self, point, max_dist, skip_last_face=False):
        # build kd tree from sliced edges - without endings (eg. only mid point)
        if not self.edge_midpoints_kd:
            self.edge_midpoints_kd = self.update_edge_midpoints_kdtree(self.bm, max_dist)

        if skip_last_face:
            def filter_func(i):
                return i != self.strip_perp_edges[-1].index
            location, index, distance = self.edge_midpoints_kd.find(point, filter=filter_func) # XXX: sometimes filter bugs out..
        else:
            location, index, distance = self.edge_midpoints_kd.find(point)

        if location:
            best_e = self.bm.edges[index]
            if distance < max_dist/2: # if very   close - ignore perpendicular part
                return best_e

            elif distance < max_dist:
                is_perpendicular_to_tan = self.is_perp( (best_e.verts[1].co - best_e.verts[0].co).normalized(), self.tangent)
                if is_perpendicular_to_tan:
                    return best_e

        return None


    def kd_tree_from_edges(self, boundary_edges, max_dist) -> None:
        ''' build kd tree from sliced edges - with endings (eg. only e.verts[0] and e.verts[1])'''
        kd_map = {}
        kd_cnt = 0
        for ei, e in enumerate(boundary_edges):
            v0, v1 = e.verts[0].co, e.verts[1].co
            edge_length = (v1 - v0).length
            n = max(1, round(edge_length / max_dist))  # Ensure at least one slice
            kd_map[ei] = [v0.lerp(v1, i / n) for i in range(n + 1)]
            kd_cnt += (n + 1)

        kd = kdtree.KDTree(kd_cnt)
        for ei, pos_list in kd_map.items():
            for pos in pos_list:
                kd.insert(pos, ei)
        kd.balance()
        return kd


    def find_closest_track_edge(self, point, max_dist, tangent, from_center_vec):
        ''' find edges max_dist from point, that are parallel to tangent and their edge.face outward_dir matches from_center_vec'''
        # use intersect_point_line - to find if point is on edge...
        passed_verts_layer = self.bm.verts.layers.int.get('passed_verts') # 0 not tracked yet, 1 tracked, 2 finished
        tracked_edges_layer = self.bm.edges.layers.int.get("tracked_edges")
        boundary_edges = [e for e in self.bm.edges
            if e.is_boundary
            and e[tracked_edges_layer] < 2   # 0- not tracked, 1 - tracked, 2 - finished tracking
            and not (e.verts[1][passed_verts_layer] and e.verts[0][passed_verts_layer]) # both verts are not passed already
            and self.is_parallel((e.verts[1].co - e.verts[0].co).normalized(), tangent)
        ] # bot verts are not passed already


        kd = self.kd_tree_from_edges(boundary_edges, max_dist)

        best_dist = 900.0
        best_edge = None

        processed_ids = set()

        for (location, index, distance) in kd.find_range(point, max_dist):
            if index in processed_ids:
                continue
            else:
                processed_ids.add(index)

            edge = boundary_edges[index]

            edge_hit, factor = intersect_point_line(point, edge.verts[0].co, edge.verts[1].co)
            if not edge_hit:
                continue

            hit_clamped = edge.verts[0].co.lerp(edge.verts[1].co, clamp(factor)) # factor   can be outside 0-1 range (relative to v0 - v1 verts)
            dist = (hit_clamped - point).length
            if dist > max_dist:
                continue
            # sort edge pts along tangent dir
            # sorted_verts = sorted(edge.verts, key=lambda vert: vert.co.dot(tangent))

            if not self.is_parallel((edge.verts[1].co - edge.verts[0].co).normalized(), tangent):
                continue

            if dist < best_dist:
                best_dist = dist
                best_edge = edge

        if not best_edge:
            return None

        if best_edge[tracked_edges_layer] != 2: # 0 - not tracked, 1 - tracked, 2 - finished tracking (do not track again..)
            # print(f"Best edge for point {point} is {best_edge.index} with distance {best_dist}")

            # calc outward_dir: out of face direction - based on link_loops
            e_dir_oriented = (best_edge.link_loops[0].vert.co - best_edge.link_loops[0].link_loop_next.vert.co).normalized()
            best_edge.link_faces[0].normal
            edge_outward_dir = best_edge.link_faces[0].normal.cross(e_dir_oriented).normalized()
            # print(f"Best edge {best_edge.index} outward dir {edge_outward_dir}")
            # see if it points towards the point or away
            are_aligned = from_center_vec.dot(edge_outward_dir)
            # print(f"{are_aligned=}")

            if are_aligned > 0: # bot vector point outward of hit_pos  - edge points away from mouse..
                return None
            return best_edge
        else:
            # print(f"Best edge for point {point} is {best_edge.index} was already tracked")
            return None


    @staticmethod
    def is_perp(closest_edge_dir, last_edge_dir, threshold=0.707) -> None:
        #    cos(0) = 1
        #    cos(20) = 0.939
        #    cos(25) = 0.906
        #    cos(30) = 0.866
        #    cos(45) = 0.707
        #    cos(60) = 0.5
        #    cos(90) = 0
        return abs(last_edge_dir.dot(closest_edge_dir)) < threshold # dot < 0.707  => the angle is between 45 and 90 degrees

    @staticmethod
    def is_parallel(closest_edge_dir, last_edge_dir, threshold=0.866) -> None:
        # by default 30 degreel and less is parallel  (cos(0-30) > 0.86)
        return abs(last_edge_dir.dot(closest_edge_dir)) > threshold

    def update_auto_spacing_factor(self, edge):
        if self.auto_spacing_factor and edge.is_boundary: # if quad then get is with (based on root_edge and opposite edge dist)
            face = edge.link_faces[0]
            face_e_cnt = len(face.edges)
            if face_e_cnt == 4:
                opposite_edge = edge.link_loops[0].link_loop_next.link_loop_next.edge
                dist = (calc_edge_center(edge) - calc_edge_center(opposite_edge)).length
                # width =  spacing_factor * 2rad
                # dist = 2 * radius * spacing_factor  => spacing_factor = dist / (2 * radius)
            elif face_e_cnt == 3: # get from opposite vert
                opposite_vert = edge.link_loops[0].link_loop_next.link_loop_next.vert
                dist = (calc_edge_center(edge) - opposite_vert.co).length
            else:
                return
            self.spacing_factor = max(dist / (2 * self.radius), 0.2 * self.spacing_factor) # prevent tiny width



    def was_vert_passed(self, vert_1, tangent) -> None:
        dist = distance_point_to_plane(self.hit_local, vert_1.co, tangent)
        return  dist > 0


    def normalize_quad_len(self, prev_edge_center, seg_width):
        # normalize seg length - around seg_width =  2*radius. - by moving hit_local closer/further from prev_edge_center
        pre_to_hit_len = (self.hit_local - prev_edge_center).length

        # scale hit_local so that  seg_width is matched
        return prev_edge_center.lerp(self.hit_local, seg_width/pre_to_hit_len) # fade hit to to 2*radius len


    @staticmethod
    def rotate_edge(edge_to_align, ref_edge, shared_face, factor=0.5):
        """
        Rotate edge defined by v1 and v2 around its midpoint to align with the edge defined by v3 and v4.

        Parameters:
        v1, v2, v3, v4: Vertices with .co attribute representing their coordinates.
        factor: A float between 0 and 1 where 0 means no rotation and 1 means full rotation.
        """
        # NOTE: what if edge has 2 link loops (linked to 2 faces)...   add shared_face check?  to be sure that edge loop belongs to the same face
        edge_align_loop = edge_to_align.link_loops[0] if len(edge_to_align.link_loops) == 1 or edge_to_align.link_loops[0].face == shared_face else edge_to_align.link_loops[1]
        v1,v2 = edge_align_loop.vert, edge_align_loop.link_loop_next.vert

        ref_edge_loop = ref_edge.link_loops[0] if len(ref_edge.link_loops) == 1 or ref_edge.link_loops[0].face == shared_face else ref_edge.link_loops[1]
        v3,v4 = ref_edge_loop.link_loop_next.vert, ref_edge_loop.vert # reverse order

        # Calculate vec_a and vec_b
        vec_a = v2.co - v1.co
        vec_b = v4.co - v3.co

        # Compute the rotation difference
        rotation_diff = vec_a.rotation_difference(vec_b)

        # Calculate the midpoint of the edge
        midpoint = (v1.co + v2.co) / 2

        faded_rot = identity_quat.slerp(rotation_diff, factor)

        # Rotate v1 and v2 around the midpoint
        v1.co = faded_rot @ (v1.co - midpoint) + midpoint
        v2.co = faded_rot @ (v2.co - midpoint) + midpoint

    def snap_to_center_if_mirror(self,verts, threshold):
        if self.use_mirror:
            for v in verts:
                if abs(v.co.x) < threshold:
                    v.co.x = 0


    def face_normal_flip(self, new_face):
        new_face.normal_update()
        if self.normal.dot(new_face.normal) < 0:
            new_face.normal_flip()
            new_face.normal_update()


    def quads_draw(self, active_obj, last_segment=False):
        if not self.hit_local:
            return

        if not self.prev_segment_hit_pos:
            self.prev_segment_hit_pos = self.hit_local.copy()
            return

        radius = self.radius / (active_obj.scale.length/math.sqrt(3)) # normalize world space radius, to object space
        mirror_snap_dist = 0.4*radius
        seg_width = 2 * radius * self.spacing_factor

        self.weighted_drag_smooth(active_obj, self.hit_local, 0.7*radius)

        for e in self.strip_perp_edges:
            e.select = False

        tangent = (self.hit_local - self.drag_co).normalized()
        self.tangent = tangent
        bi_tan = tangent.cross(self.normal).normalized()

        tracked_edges_layer = self.bm.edges.layers.int.get("tracked_edges")
        passed_verts_layer = self.bm.verts.layers.int.get('passed_verts') # 0 not tracked yet, 1 tracked, 2 finished

        #################################################################
        #####   Last  Segment ###########################################
        #################################################################

        if last_segment and self.strip_perp_edges:  # on finish glue  strip_perp_edges[-1] and closest edge
            log("Last segment")
            last_strip_edge = self.strip_perp_edges[-1]
            # detect if last strip_per_edge - belogns to hole (tri or quad) if so - special case and just fill and finish

            hole_edges = detect_hole_adjacent_to_edge(last_strip_edge)
            if hole_edges:
                log("Hole detected", hole_edges)
                bmesh.ops.contextual_create(self.bm, geom=hole_edges) # fill hole
                bmesh.update_edit_mesh(active_obj.data) # or else crash
                return


            closest_edge  = self.find_closest_boundary_edge(self.hit_local, radius, skip_last_face=True)
            if not closest_edge:
                # check if we tracked edges - if so - finish them
                if self.track_border_edge1 and self.track_border_edge2:
                    sorted_verts = sorted(self.track_border_edge1.verts, key=lambda vert: vert.co.dot(tangent))
                    v3 = sorted_verts[1]
                    sorted_verts = sorted(self.track_border_edge2.verts, key=lambda vert: vert.co.dot(tangent))
                    v4 = sorted_verts[1]
                    v1, v2 = last_strip_edge.verts

                    # untwist
                    v1_v2_dir = (v2.co - v1.co).normalized()
                    v3_v4_dir = (v4.co - v3.co).normalized()
                    if v1_v2_dir.dot(v3_v4_dir) < 0:
                        v3, v4 = v4, v3
                    new_face = self.bm.faces.new([v1, v2, v4, v3])
                    self.face_normal_flip(new_face)
                    bmesh.update_edit_mesh(active_obj.data) # or else crash

                return

            # compare dot of last edge and current edge - if perpendicular then skip
            closest_edge_dir = (closest_edge.verts[1].co - closest_edge.verts[0].co).normalized()

            is_perpendicular = self.is_perp(closest_edge_dir, tangent)
            if is_perpendicular and last_strip_edge != closest_edge: # create quad between last segment hit and closest edge  using bmesh.ops.contextual_create(bm,

                # NOTE: check if closest_edge and prev_last_edge are connected - share same vert
                closest_edge_verts = set(closest_edge.verts)
                prev_last_edge_verts = set(last_strip_edge.verts)
                edges_share_vert = closest_edge_verts.intersection(prev_last_edge_verts) # they probably form triangle

                if edges_share_vert:
                    non_shared_verts = list(prev_last_edge_verts.symmetric_difference(closest_edge_verts))
                    if len(self.strip_perp_edges)>1: # weld non shared verts together if this is finish of strip from at least 2 edges
                        # sort by tan
                        sorted_shared_verts = sorted(non_shared_verts, key=lambda vert: vert.co.dot(tangent))
                        verts_to_merge_map = {sorted_shared_verts[0]: sorted_shared_verts[1]}
                        bmesh.ops.weld_verts(self.bm, targetmap=verts_to_merge_map)

                        del self.strip_perp_edges[-1] # became welded
                        if len(self.strip_perp_edges) > 1 and self.strip_perp_edges[-1].is_valid: # do not smooth first strip (root) edge
                            laplacian_smooth(self.strip_perp_edges[-1].verts)
                        if closest_edge.is_valid: # could be removed by weld
                            closest_edge.select = True

                        bmesh.update_edit_mesh(active_obj.data) # or else crash
                        return
                    else: # fill with triangle
                        tri_vers = closest_edge_verts.union(prev_last_edge_verts)
                        new_face = self.bm.faces.new(tri_vers)
                        self.face_normal_flip(new_face)

                        bmesh.update_edit_mesh(active_obj.data) # or else crash
                        return


                can_delete = len(last_strip_edge.verts[0].link_edges) <= 2 and len(last_strip_edge.verts[1].link_edges) <= 2
                # was_prev_edge_from_track = last_strip_edge[tracked_edges_layer] == 2


                v1, v2 = last_strip_edge.verts
                v3, v4 = closest_edge.verts

                # untwist
                v1_v2_dir = (v2.co - v1.co).normalized()
                v3_v4_dir = (v4.co - v3.co).normalized()
                if v1_v2_dir.dot(v3_v4_dir) < 0:
                    v3, v4 = v4, v3

                new_face = self.bm.faces.new([v1, v2, v4, v3])
                self.face_normal_flip(new_face)

                last_edges_dist = (calc_edge_center(closest_edge) - self.prev_segment_hit_pos).length
                if last_edges_dist < 1.4*radius and can_delete: # move last edge points to closest edge - and weld
                    bmesh.ops.dissolve_edges(self.bm, edges=[last_strip_edge], use_verts=True)
                    del self.strip_perp_edges[-1]

                # bmesh.ops.smooth_laplacian_vert( # does nothing..
                # bmesh.ops.smooth_vert(s  # works but not usable (shinkage))
                one_quad = len(self.strip_perp_edges) == 1  # looks like we closed single quad hole
                # smooth_verts = [v for v in self.strip_perp_edges[-1].verts if (v[passed_verts_layer] == 0 or one_quad) and v.is_valid]  # do not smoooth passed  track verts
                # laplacian_smooth(smooth_verts)
                if one_quad:
                    self.rotate_edge(self.strip_perp_edges[-1], closest_edge, new_face, factor=0.3)

                closest_edge.select = True
                self.rotate_edge(closest_edge, self.strip_perp_edges[-1], new_face) # rotate v12 toward v34

                bmesh.update_edit_mesh(active_obj.data) # or else crash
            return




        #################################################################
        #####   TRACKING PART ###########################################
        #################################################################

        log("\n")
        # NOTE:  do substep check for adjacent parallel edges - we will trakc them till they end and then glue to it. Unless they become perpendicular, or too far...

        global SIDE_PTS,  TRACK_PTS
        search_rad = lerp(radius, seg_width, 0.2) # weighted search radius 80% of rad
        pos1 =  self.hit_local + bi_tan * radius
        edge_candidate1 = self.find_closest_track_edge(pos1, search_rad, tangent, bi_tan)

        pos2 =  self.hit_local - bi_tan * radius
        edge_candidate2 = self.find_closest_track_edge(pos2, search_rad, tangent, -bi_tan)

        SIDE_PTS = [pos1, pos2]

        # if both tracked edges are same - pick the one closest to pos1 (or pos2)
        if edge_candidate1 and edge_candidate2 and edge_candidate1 == edge_candidate2:
            log(f"\tBoth edges are same: {edge_candidate1.index}")
            e_center = calc_edge_center(edge_candidate1)
            dist1  = (e_center - pos1).length
            dist2  = (e_center - pos2).length
            if dist1 < dist2:
                log("\tClearing e2")
                edge_candidate2 = None
            else:
                log("\tClearing e1")
                edge_candidate1 = None


        vert1_passed = False
        vert_1 = None
        if edge_candidate1:
            sorted_verts = sorted(edge_candidate1.verts, key=lambda vert: vert.co.dot(tangent))

            # fix corner  cases
            # case1: we  did not track - and got near new edge1.verts - if we passed aleady v0 - we skip v0 quad gen..
            if self.track_border_edge1 is None: # we must have just got near new edge... check if we already passed v0 (or v1?)
                if self.was_vert_passed(sorted_verts[0], tangent):
                    sorted_verts[0][passed_verts_layer] = True
                if self.was_vert_passed(sorted_verts[1], tangent):
                    sorted_verts[1][passed_verts_layer] = True

            # case2: - we tracked e1,  but switched to e2 - but e2.verts are bot already passed - detect, and mark for skip
            elif self.track_border_edge1 != edge_candidate1:
                if self.was_vert_passed(sorted_verts[0], tangent) and self.was_vert_passed(sorted_verts[1], tangent):
                    sorted_verts[0][passed_verts_layer] = True
                    sorted_verts[1][passed_verts_layer] = True


            if not sorted_verts[0][passed_verts_layer]: # not passed vert? use first
                vert_1 = sorted_verts[0]
            elif not sorted_verts[1][passed_verts_layer]: # not passed vert? use second
                vert_1 = sorted_verts[1]

            if vert_1:
                if self.was_vert_passed(vert_1, tangent):
                    vert1_passed = True
                log(f"TRACK1: {edge_candidate1.index}, State:{edge_candidate1[tracked_edges_layer]}, Vert [{vert_1.index}] Passed: {vert1_passed}")
            else: # mark whole edge as already passed (happens only on self.first_step ? )
                edge_candidate1[tracked_edges_layer] = 2
                edge_candidate1 = None
        else:
            log("TRACK1: None")

        self.track_border_edge1 = edge_candidate1

        vert2_passed = False
        vert_2 = None
        if edge_candidate2:
            sorted_verts = sorted(edge_candidate2.verts, key=lambda vert: vert.co.dot(tangent))

            if self.track_border_edge2 is None: # we must have just got near new edge... check if we already passed v0 (v1 too?)
                if self.was_vert_passed(sorted_verts[0], tangent):
                    sorted_verts[0][passed_verts_layer] = True
                if self.was_vert_passed(sorted_verts[1], tangent):
                    sorted_verts[1][passed_verts_layer] = True

            # case2: - we tracked e1,  but switched to e2 - but e2.verts are bot already passed - detect, and mark for skip
            elif self.track_border_edge2 != edge_candidate2:
                if self.was_vert_passed(sorted_verts[0], tangent) and self.was_vert_passed(sorted_verts[1], tangent):
                    sorted_verts[0][passed_verts_layer] = True
                    sorted_verts[1][passed_verts_layer] = True

            if not sorted_verts[0][passed_verts_layer]: # not passed vert? use first
                vert_2 = sorted_verts[0]
            elif not sorted_verts[1][passed_verts_layer]: # not passed vert? use second
                vert_2 = sorted_verts[1]

            if vert_2:
                if self.was_vert_passed(vert_2, tangent):
                    vert2_passed = True
                log(f"TRACK2: {edge_candidate2.index}, State:{edge_candidate2[tracked_edges_layer]}, Vert [{vert_2.index}] Passed: {vert2_passed}")
            else: # mark whole edge as already passed (happens only on self.first_step ? )
                edge_candidate2[tracked_edges_layer] = 2
                edge_candidate2 = None

        else:
            log("TRACK2: None")

        self.track_border_edge2 = edge_candidate2

        ####  TRACKING QUAD ADD ############################################################################################

        if vert1_passed or vert2_passed:
            log("Track ended: quad draw")
            if self.strip_perp_edges:
                v1, v2 = self.strip_perp_edges[-1].verts[0], self.strip_perp_edges[-1].verts[1]
            else: # possibly floating  edge...
                v1 = self.bm.verts.new(self.prev_segment_hit_pos - bi_tan * radius)
                v2 = self.bm.verts.new(self.prev_segment_hit_pos + bi_tan * radius)
                self.snap_to_center_if_mirror((v1,v2), mirror_snap_dist)

            avg_tracked_edge_len = 0.0
            if vert_1:
                v3 = vert_1
                vert_1[passed_verts_layer] = True # mark for ignore
                edge_candidate1[tracked_edges_layer] += 1 # mark and bump to tracked - 1, or finished - 2
                avg_tracked_edge_len = edge_candidate1.calc_length()
            else:
                # TODO: maybe  should search for closest vert (without using rail edges )?
                v3 = self.bm.verts.new(self.hit_local + bi_tan * radius)
                self.snap_to_center_if_mirror((v3,), mirror_snap_dist)


            if vert_2:
                v4 = vert_2
                vert_2[passed_verts_layer] = True # mark for ignore
                edge_candidate2[tracked_edges_layer] += 1 # mark and bump to tracked - 1, or finished - 2
                avg_tracked_edge_len = (avg_tracked_edge_len + edge_candidate2.calc_length()) / 2 if vert_1 else edge_candidate2.calc_length()
            else:
                # TODO: maybe  should search for closest vert (without using rail edges )?
                v4 = self.bm.verts.new(self.hit_local - bi_tan * radius) # we should use self.edges_strip[-1]
                self.snap_to_center_if_mirror((v4,), mirror_snap_dist)


            self.bm.verts.ensure_lookup_table()
            log(f"{v1.index=}, {v2.index=}, {v3.index=}, {v4.index=}")

            # check if v3  and 4 need to swap (copare dot of v1-v2 and v3-v4))
            v1_v2_dir = (v2.co - v1.co).normalized()
            v3_v4_dir = (v4.co - v3.co).normalized()
            if v1_v2_dir.dot(v3_v4_dir) < 0:
                v3, v4 = v4, v3

            new_face = self.bm.faces.new([v1, v2, v4, v3])
            self.face_normal_flip(new_face)

            new_edge = self.bm.edges.get((v3, v4))

            # mark all face edges as done
            for e in new_face.edges:
                e[tracked_edges_layer] = 2

            edge_v12 = self.bm.edges.get((v1, v2)) # opposite edge to new one on new_face
            strip_edges_cnt = len(self.strip_perp_edges)
            if strip_edges_cnt == 0: # we draw first quad - check if it is not very narrow - if so - delete it
                floating_edge = len(edge_v12.verts[0].link_edges) <= 2 and len(edge_v12.verts[1].link_edges) <= 2
                if  floating_edge: # must be only linked only to new face
                    # if new_face_width < 0.5*seg_width: # annoyig - since with can be dynamic. Use tracked edge length instead
                    new_face_width = (calc_edge_center(new_edge) - self.prev_segment_hit_pos).length
                    if new_face_width < 0.5*avg_tracked_edge_len:
                        self.bm.verts.remove(v1)
                        self.bm.verts.remove(v2)
                        # self.bm.edges.remove(edge_v12)  # doesn not remove verts
                        new_face = None

            elif strip_edges_cnt == 1:  # looks like we drawn 1 quad:
                # just do align prev edge to new edge
                self.rotate_edge(self.strip_perp_edges[-1], new_edge, new_face) # rotate v12 toward v34

            else:  # check if previous strip loop can be dissolved - if too close to new edge
                strip_type = len(edge_v12.verts[0].link_edges) == 3 and len(edge_v12.verts[1].link_edges) == 3 # two T junctions - just strip - that can be dissolved..
                strip_type_cond2 = len(edge_v12.verts[0].link_faces) == 2 and len(edge_v12.verts[1].link_faces) == 2
                if strip_type and strip_type_cond2 : # prev edge was strip - if too narow - dissolve
                    new_face_width = (calc_edge_center(new_edge) - self.prev_segment_hit_pos).length
                    if new_face_width < 0.7*seg_width:
                        log(f"Dissolving strip {edge_v12=} #########################################")
                        bmesh.ops.dissolve_edges(self.bm, edges=[edge_v12], use_verts=True)
                        del self.strip_perp_edges[-1] # remove last edge from strip
                        # XXX: run  e.is_vaild ? and clean dissolved edges from lists?
                        new_face = None

                self.update_auto_spacing_factor(new_edge)

            self.strip_perp_edges.append(new_edge)
            if new_face:
                for fe in new_face.edges:
                    fe[tracked_edges_layer] = 2 # do not track

            bmesh.update_edit_mesh(active_obj.data)
            self.bm.edges.ensure_lookup_table()

            if self.split_undo:
                bpy.ops.ed.undo_push(message="Draw Xray Quad")
            self.edge_midpoints_kd = None # reset kd tree
            self.prev_segment_hit_pos = self.hit_local.copy()
            return

        # if vert_1 or vert_2: # do not fill quads if we are following vert..
        #     return

        if self.track_border_edge1 or self.track_border_edge2: # do not create new quad if we are still tracking at least one edge
            return

        #################################################################
        ##### END - TRACKING PART #######################################
        #################################################################



        seg_len = (self.hit_local - self.prev_segment_hit_pos).length
        log(f"{seg_len=}")
        if seg_len < seg_width: # just continue drawing
            return


        #################################################################
        ##### Floating Quads Part #######################################
        #################################################################

        new_face = None
        if self.first_step:
            log("First step")
            # closest_edge = self.find_closest_boundary_edge(self.prev_segment_hit_pos, radius)

            closest_edge = self.strip_perp_edges[-1] if self.strip_perp_edges else None

            # is_perpendicular = False
            # if closest_edge:
            #     closest_edge_dir = (closest_edge.verts[1].co - closest_edge.verts[0].co).normalized()
            #     is_perpendicular = self.is_perp(closest_edge_dir, bi_tan)

            # if closest_edge and not is_perpendicular: # glue to existing edge
            if closest_edge: # glue to existing edge
                v1, v2 = closest_edge.verts
                prev_edge_center = (v1.co + v2.co) / 2

                normalized_hit_local = self.normalize_quad_len(prev_edge_center, seg_width)
                v4 = self.bm.verts.new(normalized_hit_local + bi_tan * radius )
                v3 = self.bm.verts.new(normalized_hit_local - bi_tan * radius )


                v1_v2_dir = (v2.co - v1.co).normalized()
                v3_v4_dir = (v4.co - v3.co).normalized()

                # use dot product to determine if the new edge is in the same direction as the existing edge and if v1 and v2 need to swap
                if v1_v2_dir.dot(v3_v4_dir) < 0:
                    v3, v4 = v4, v3

                new_face = self.bm.faces.new((v1, v2, v4, v3))
                self.face_normal_flip(new_face)

                for e in new_face.edges: # do not track new face edges - fixes fan edges bug
                    e[tracked_edges_layer] = 2

                # NOTE: fade first edge into second edge dir
                # Calculate vec_a and vec_b
                new_edge = self.bm.edges.get((v3, v4))
                self.rotate_edge(closest_edge, new_edge, new_face) # rotate v12 toward v34


                self.snap_to_center_if_mirror(new_edge.verts , mirror_snap_dist)
                self.strip_perp_edges.append(closest_edge)
                self.strip_perp_edges.append(new_edge)


            else: # just add 'floating' quad
                # Create quad vertices for the first stroke (original code)
                v1 = self.hit_local + bi_tan * radius
                v2 = self.hit_local - bi_tan * radius
                v3 = self.prev_segment_hit_pos - bi_tan * radius
                v4 = self.prev_segment_hit_pos + bi_tan * radius

                new_verts = [self.bm.verts.new(v) for v in [v1, v2, v3, v4]]
                new_face = self.bm.faces.new(new_verts)
                self.face_normal_flip(new_face)

                for e in new_face.edges: # do not track new face edges - fixes fan edges bug
                    e[tracked_edges_layer] = 2

                self.snap_to_center_if_mirror(new_verts, mirror_snap_dist)
                self.strip_perp_edges.append(self.bm.edges.get((new_verts[2], new_verts[3])))
                self.strip_perp_edges.append(self.bm.edges.get((new_verts[0], new_verts[1]))) # last edge

            self.first_step = False

        elif self.strip_perp_edges: #  Extrude strip
            log("Extrude step")
            ret = bmesh.ops.extrude_edge_only(self.bm, edges=[self.strip_perp_edges[-1]])
            geom = ret["geom"]
            new_verts = [ele for ele in geom if isinstance(ele, bmesh.types.BMVert)]
            new_face = next((ele for ele in geom if isinstance(ele, bmesh.types.BMFace)), None)
            new_edge = self.bm.edges.get((new_verts[0], new_verts[1]))

            for e in new_face.edges: # do not track new face edges - fixes fan edges bug
                e[tracked_edges_layer] = 2

            # print(f"{self.last_edge[-1].index=}")
            # print(f"{new_edge.index=}")

            # Move the new vertices to the current position
            # bmesh.ops.translate(self.bm, vec=step, verts=new_verts)

            v1 = new_face.verts[-3]
            v2 = new_face.verts[-4]

            v3 = new_face.verts[-2] # same as new edge verts..
            v4 = new_face.verts[-1]

            # print(f"{v1.index=}, {v2.index=}, {v3.index=}, {v4.index=}")

            prev_edge_center = self.prev_segment_hit_pos    # sames as - (v1.co + v2.co) / 2

            tip_hit_local = self.normalize_quad_len(prev_edge_center, seg_width)

            # prevent twist of face
            v1_v2_dir = (v2.co - v1.co).normalized()
            if bi_tan.dot(v1_v2_dir) < 0:
                bi_tan = -bi_tan

            v3.co = tip_hit_local - bi_tan * radius
            v4.co = tip_hit_local + bi_tan * radius


            # makes sure v3 and 4 on tangent is extended/shrink toward closest verts..  (not used - we have tan gracking now)
            # self.bm.verts.ensure_lookup_table()
            # ext_tan_to_neighbor_verts = []
            # for new_vert in [v3, v4]:
            #     co, closest_v_id, dist = self.bm_verts_kdtree.find(new_vert.co)
            #     if dist < radius:
            #         close_vert = self.bm.verts[closest_v_id]
            #         prev_to_closest_vert = close_vert.co - prev_edge_center
            #         close_vert_on_tan = prev_to_closest_vert.project(tangent)
            #
            #         diff_on_tan = close_vert_on_tan - (tip_hit_local - prev_edge_center)
            #         ext_tan_to_neighbor_verts.append(diff_on_tan)
            #
            # offset_cnt = len(ext_tan_to_neighbor_verts)
            #
            # if offset_cnt == 2: # average offset
            #     offset_averg = (ext_tan_to_neighbor_verts[0] + ext_tan_to_neighbor_verts[1]) / 2
            #     v3.co += offset_averg
            #     v4.co += offset_averg
            # elif offset_cnt == 1: # both verts extend same distance
            #     v3.co += ext_tan_to_neighbor_verts[0]
            #     v4.co += ext_tan_to_neighbor_verts[0]
            # if offset_cnt > 0:
            #     tip_hit_local = (v3.co + v4.co) / 2 # update from above..



            # prevet flip when using fade..
            v3_v4_dir = (v4.co - v3.co).normalized()
            # if v1_v2_dir.dot(v3_v4_dir) < 0:
            #     print("flip")
            #     v3, v4 = v4, v3

            prev_edge_len = self.strip_perp_edges[-1].calc_length()

            # calc aver dir by extruded face 'rail' edges verts
            v1_aver_co = (v3.co + v1.co) / 2
            v2_aver_co = (v4.co + v2.co) / 2

            # fade prev edge dir into perpendicular by 0.5
            aver_dir = (v1_aver_co - v2_aver_co).normalized()

            v2.co = prev_edge_center - aver_dir * (prev_edge_len / 2)
            v1.co = prev_edge_center + aver_dir * (prev_edge_len / 2)

            #  calc face normal, and compare to self.normal - if not dot then flip
            self.face_normal_flip(new_face)
            self.snap_to_center_if_mirror(new_edge.verts, mirror_snap_dist)
            self.strip_perp_edges.append(new_edge)


        self.strip_perp_edges[-1].select = True
        for fe in new_face.edges:
            fe[tracked_edges_layer] = 2 # do not track

        # Update mesh
        bmesh.update_edit_mesh(active_obj.data)
        self.bm.edges.ensure_lookup_table()

        if self.split_undo:
            bpy.ops.ed.undo_push(message="Draw Xray Quad")

        self.edge_midpoints_kd = None # reset midpoins kd tree - to force update
        self.prev_segment_hit_pos = self.hit_local.copy()


    def adjust_size(self, context, event):
        region = context.region
        rv3d = context.region_data
        curr_loc3d_from_2d = view3d_utils.region_2d_to_location_3d(region, rv3d, (event.mouse_region_x, 0), self.hit_world)
        start_loc3d_from_2d = view3d_utils.region_2d_to_location_3d(region, rv3d, (self.f_start_xloc, 0), self.hit_world)
        diff = (curr_loc3d_from_2d - start_loc3d_from_2d).length
        if self.f_start_xloc > event.mouse_region_x:
            diff *= -1
        tool = context.workspace.tools.from_space_view3d_mode(context.mode)
        min_size = self.radius * 0.01
        mit_size_first_dec = round_to_first_digit(min_size)
        tool_props = tool.operator_properties("mesh.quad_xray_brush")
        rad = max(mit_size_first_dec, self.f_start_rad+diff)
        tool_props.radius = rad # cos self.radius  is not updated when changes from workspace tool...vol
        self.radius = rad #now mostly for UI drawing

    def get_hit(self, context, event):
        mouse_2d_co = event.mouse_region_x, event.mouse_region_y
        self.hit_local, normal = raycast(context, mouse_2d_co)
        self.normal = normal # for quad drawing

        if not self.resize or not self.hit_world:
            mat = context.active_object.matrix_world
            if self.hit_local:
                self.hit_world = mat @ self.hit_local
            else:
                rv3d = context.region_data
                depth = self.hit_world if self.hit_world else Vector((0, 0, 0))
                location_from_2d = view3d_utils.region_2d_to_location_3d(context.region , rv3d, mouse_2d_co, depth)
                self.hit_world = location_from_2d
                self.hit_local = mat.inverted() @ location_from_2d
                normal = get_view_vec(context)
                self.normal = normal
            quat = normal.to_track_quat('Z','Y')  # to_track_quat(track, up)
            rot_mat = quat.to_matrix().to_4x4()

        else:
            # get rot from view matrix
            view_vec = get_view_vec(context)
            rot_mat = view_vec.to_track_quat('Z','Y').to_matrix().to_4x4()

        rad_scale_mat = Matrix.Scale(self.radius, 4)
        self.matrix_basis = Matrix.Translation(self.hit_world) @ rot_mat @ rad_scale_mat # for gizmoe



    def finalize(self, context):
        self.resize = False
        context.scene.xray_gizmo_show = True
        wm = context.window_manager
        wm.gizmo_group_type_ensure("MESH_GGT_brush_circle_xray")  # link back gizmo - brush outline
        bpy.types.SpaceView3D.draw_handler_remove(self._circel2d_handle, 'WINDOW')
        if self._ui3d_handle:
            bpy.types.SpaceView3D.draw_handler_remove(self._ui3d_handle, 'WINDOW')

        close_bmesh(context, self.bm, context.active_object.data)
        # if context.scene.tool_settings.use_snap:
        #     bpy.ops.transform.translate(snap=True)

        tool = context.workspace.tools.from_space_view3d_mode(context.mode)
        tool_props = tool.operator_properties("mesh.quad_xray_brush")
        if self.auto_radius: # or else radius wont be save
            tool_props.radius = self.radius
        if self.auto_spacing_factor:
            tool_props.spacing_factor = self.spacing_factor # or else width factor wont be saved

        global SIDE_PTS, TRACK_PTS, HOVER_EDGE
        SIDE_PTS.clear()
        TRACK_PTS.clear()
        HOVER_EDGE = None

        context.area.tag_redraw()

    def hover_highlight_closest_edge(self, obj):
        global HOVER_EDGE
        if not self.strip_perp_edges: # we did not draw any quad yet..
            HOVER_EDGE = None
            return

        radius = self.radius / (obj.scale.length/math.sqrt(3)) # normalize world space radius, to object space
        closest_edge  = self.find_closest_boundary_edge(self.hit_local, radius, skip_last_face=True) # TODO: use cached kd-tree if possible.. on mouse move...
        if not closest_edge:
            HOVER_EDGE = None
            return

        # tangent = (self.hit_local - self.drag_co).normalized()

        last_strip_edge = self.strip_perp_edges[-1]
        # compare dot of last edge and current edge - if perpendicular then skip
        closest_edge_dir = (closest_edge.verts[1].co - closest_edge.verts[0].co).normalized()

        # is_perpendicular = self.is_perp(closest_edge_dir, tangent)
        if last_strip_edge != closest_edge: # create quad between last segment hit and closest edge  using bmesh.ops.contextual_create(bm,
            HOVER_EDGE = closest_edge
        else:
            HOVER_EDGE = None


    def modal(self, context, event):
        try:
            obj = context.active_object
            if event.type == 'MOUSEMOVE':
                self.get_hit(context, event) # sets new self.hit_loc
                self.hover_highlight_closest_edge(obj)
                if self.strip_perp_edges and len(self.strip_perp_edges[-1].link_faces) > 1: # last edge must have collapsed hole
                    self.finalize(context)
                    return {'FINISHED'}
                if self.resize:
                    self.adjust_size(context, event)
                else:
                    self.quads_draw(obj)

                context.area.tag_redraw()

            elif event.type == 'LEFTMOUSE':
                if event.value == 'PRESS' and not self.resize : # dealt in WorkTool keybindings
                    context.scene.xray_gizmo_show = True

                if event.value == 'RELEASE':
                    if not self.resize:
                        self.get_hit(context, event) # sets new self.hit_loc
                        self.quads_draw(obj, last_segment=True)

                    self.finalize(context)
                    return {'FINISHED'}

            elif event.type == 'RIGHTMOUSE' and event.value == 'PRESS':
                tool = context.workspace.tools.from_space_view3d_mode(context.mode)
                tool_props = tool.operator_properties("mesh.quad_xray_brush")
                if self.resize:
                    tool_props.radius =  self.reset_rad

                self.finalize(context)
                return {'CANCELLED'}

        except Exception as e:
            self.report({'ERROR'}, 'Exception occurred while running "Draw Quads" operator. Finishing')
            print(traceback.format_exc())
            self.finalize(context)
            return {'FINISHED'}

        return {'RUNNING_MODAL'}

    @classmethod
    def poll(cls, context):
        return context.active_object and context.active_object.type == 'MESH'

    def vert_kd_tree_build(self) -> None:
        kd = kdtree.KDTree(len(self.bm.verts))
        for i, v in enumerate(self.bm.verts):
            kd.insert(v.co, i)
        kd.balance()
        return kd


    def invoke(self, context, event):
        active_obj = context.active_object

        if active_obj.type != 'MESH':
            self.report({'ERROR'}, 'Select proper mesh object for particle hair drawing. Cancelling')
            return {'CANCELLED'}

        context.scene.xray_gizmo_show = False # is replaced with modal draw function

        self.bm = bmesh.from_edit_mesh(active_obj.data)

        self.bm.select_mode = {'EDGE'}
        self.bm.select_flush_mode()
        for e in self.bm.edges:
            e.select = False
        for f in self.bm.faces:
            f.select = False
        for v in self.bm.verts:
            v.select = False
        bmesh.update_edit_mesh(active_obj.data) # update sel

        tracked_edges_layer = self.bm.edges.layers.int.get("tracked_edges")
        if not tracked_edges_layer:
            tracked_edges_layer = self.bm.edges.layers.int.new('tracked_edges') # 0 not tracked yet, 1 tracked, 2 finished
        for edge in self.bm.edges:
            edge[tracked_edges_layer] = 0

        passed_verts_layer = self.bm.verts.layers.int.get('passed_verts')
        if not passed_verts_layer:
            passed_verts_layer = self.bm.verts.layers.int.new('passed_verts') # 0 not tracked yet, 1 tracked, 2 finished
        for vert in self.bm.verts: # used for passed verts - to ignore
            vert[passed_verts_layer] = False


        self.track_border_edge1 = None
        self.trakc_vert1 = None
        self.track_border_edge2 = None
        self.track_vert1 = None


        self.strip_perp_edges = []
        self.first_step = True
        self.mouse_init_co = event.mouse_region_x, event.mouse_region_y #    for drawing brush size (not used anymore..)


        if not draw_xray.TARGET_BVH_LIST:
            depsgraph = context.evaluated_depsgraph_get()
            draw_xray.TARGET_BVH_LIST = get_obj_mesh_bvht(context, depsgraph, applyModifiers=True, world_space=True)

        if not draw_xray.TARGET_BVH_LIST: # no target was set, or target is not visible..
            self.report({'ERROR'}, 'DrawXray: snap cache not found (snap target does not exist or it is hidden). Cancelling')
            return {'CANCELLED'}

        if active_obj.data.use_mirror_x: # does nothing for now...
            self.mirror_map = get_mirrored_verts(active_obj.data)

        self.use_mirror = active_obj.data.use_mirror_x or active_obj.modifiers.get('Mirror') is not None

        self.prev_segment_hit_pos = None # for determining if quad width threshold is reached
        self.hit_local = None #raycast on object loc
        self.hit_world = None
        self.normal = Vector((0, 0, 1))

        self.get_hit(context, event) # init self.hit_loc, self.hit_world
        self.tangent = None

        if self.resize:  #f - start as brush resize oper
            self.f_start_xloc = event.mouse_region_x
            self.f_start_rad = self.radius
            self.reset_rad = self.radius
        else:
            radius = self.radius / (active_obj.scale.length/math.sqrt(3)) # normalize world space radius, to object space
            # self.bm_verts_kdtree = self.vert_kd_tree_build() # NOTE: replace with dynamic one?

            self.edge_midpoints_kd  = self.update_edge_midpoints_kdtree(self.bm, radius) # cache for fast hover

            self.drag_co = None # delayed co for calculation tangent
            if self.hit_local:
                # no tangent yet calc based on global GIZMO_MOUSE_CO and event mouse co
                if GIZMO_MOUSE_CO:
                    prev_mouse_2d = GIZMO_MOUSE_CO[0]
                    rv3d = context.region_data
                    prev_2d_to_3d = view3d_utils.region_2d_to_location_3d(context.region, rv3d, prev_mouse_2d, active_obj.matrix_world @ self.hit_local)
                    prev_local = active_obj.matrix_world.inverted() @ prev_2d_to_3d
                    self.tangent = (self.hit_local - prev_local).normalized()
                    self.drag_co = prev_local.copy()
                    self.weighted_drag_smooth(active_obj, self.hit_local, 0.7*radius)

                    root_edge = self.find_closest_boundary_edge(self.hit_local, radius)
                    if root_edge:
                        edge_dir = (root_edge.verts[0].co-root_edge.verts[1].co).normalized()
                        # is_perpendicular_to_tan = self.is_perp(edge_dir, tangent)
                        # print(f"{is_perpendicular_to_tan=} edge {root_edge}")
                        # if is_perpendicular_to_tan: # ignore - is_perpendicular (no tan yet)
                        if True:
                            root_edge.verts[0][passed_verts_layer] = True # mark as passed
                            root_edge.verts[1][passed_verts_layer] = True # mark passed
                            root_edge[tracked_edges_layer] = 2 # mark as finished

                            self.strip_perp_edges.append(root_edge)
                            root_edge[tracked_edges_layer] = 2 # do not track

                            if self.auto_radius:
                                new_rad = 0.9 * 0.5 * root_edge.calc_length()
                                self.radius = new_rad * (active_obj.scale.length/math.sqrt(3)) # 90% - gives nicer tracking..
                                self.edge_midpoints_kd  = self.update_edge_midpoints_kdtree(self.bm, new_rad) # cache for fast hover

                            if self.auto_spacing_factor: # if quad then get is with (based on root_edge and opposite edge dis
                                self.update_auto_spacing_factor(root_edge)


        # normal = get_view_vec(context)
        # view_local = active_obj.matrix_world.inverted()  @ Vector((normal.x, normal.y, normal.z, 0)) #XXX: cos it is dir. check with blender build-in raycast
        # self.view_local = view_local.to_3d()
        # screen_up = view_ma[1].xyz
        # screen_right = view_ma[1].xyz

        args = (self, (0.9,0.9,0.5,1))
        self._circel2d_handle = bpy.types.SpaceView3D.draw_handler_add(draw_circle_px, args, 'WINDOW', 'POST_VIEW')
        # if self.strengthen:
        #     self._ui3d_handle = bpy.types.SpaceView3D.draw_handler_add(draw_text_px, args, 'WINDOW', 'POST_PIXEL')
        self._ui3d_handle = bpy.types.SpaceView3D.draw_handler_add(draw_debug_px, args, 'WINDOW', 'POST_VIEW')
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}
