# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2017 JOSECONSCO
# Created by JOSECONSCO


import bpy
import bmesh
from mathutils.bvhtree import BVHTree
from bpy_extras import view3d_utils

def get_addon_name():
    return __package__.split(".")[-1]

def addon_name_lowercase():
    return get_addon_name().lower()

def get_addon_preferences():
    # return bpy.context.preferences.addons[get_addon_name()].preferences
    return bpy.context.preferences.addons[__package__].preferences


def get_collection_objects(coll, active_obj):
    col_objs = set()
    for obj in coll.all_objects: # obj from coll and child colls
        if obj.visible_get() and obj.display_type in ('SOLID','TEXTURED') and obj != active_obj: # new algo can handle everything (curves, groups, ..)
            col_objs.add(obj.name)
    # for coll in coll.children_recursive:
    #     for obj in coll.objects:
    #         if obj.type == 'MESH':
    #             col_objs.append(obj)
    return col_objs


def get_nested_obj_depsgrapth_bm(bm, depsgraph, src_objs : set, apply_modifiers=True, matrix=None):
    ''' handles everything : groups, curves etc
       Returns: None - modifies passed bm '''

    def is_object_instance_from_selected(object_instance, base_ob):
        # For instanced objects we check selection of their instancer (more accurately: check
        # selection status of the original object corresponding to the instancer).
        if object_instance.parent: #  it refers to group instance, rather than  rad array modifier.
            return object_instance.parent.original.name in base_ob
        # For non-instanced objects we check selection state of the original object.
        return object_instance.object.original.name in base_ob

    def get_obj_for_instance(object_instance, apply_modifiers):
        if apply_modifiers:
            return object_instance.object
        else:
            return object_instance.object.original

    '''Works in groups too in recursive way. apply_modifiers=False => modes get ignored competly'''
    for dep_obj_info in depsgraph.object_instances:
        if not is_object_instance_from_selected(dep_obj_info, src_objs):
            continue # We only export selected objects

        orig_obj = get_obj_for_instance(dep_obj_info, apply_modifiers)
        if not orig_obj.data or orig_obj.type != 'MESH' or orig_obj.display_type not in ('SOLID','TEXTURED'): # skip for non-geometry objects.
            continue

        # print('inter on: '+orig_obj.name)
        # bm.from_mesh(orig_obj.to_mesh())   # with modifiers
        # bm.transform(dep_obj_info.matrix_world)

        me = orig_obj.to_mesh()
        # print(f"{len(me.vertices)=}")
        me.transform(dep_obj_info.matrix_world)
        bm.from_mesh(me) # should join all meshes into one bm

        # me.to_mesh_clear()
        orig_obj.to_mesh_clear()

    if matrix:
        bm.transform(matrix)


def debug_mesh(bm):
    mesh = bpy.data.meshes.new('temp_mesh')
    bm.to_mesh(mesh)
    new_obj = bpy.data.objects.new('temp_obj', mesh)
    bpy.context.scene.collection.objects.link(new_obj)


def get_obj_mesh_bvht(context, depsgraph, applyModifiers=True, world_space=True):
    '''Target worlds space, but in active obj space. Returns BVHTree'''
    active_obj = context.active_object

    obj_xray_props = active_obj.xray_props
    scn_xray_props = context.scene.xray_props
    shared_xray_props = scn_xray_props if obj_xray_props.settings_mode == 'GLOBAL' else obj_xray_props

    if applyModifiers:
        if world_space:
            # #? wont work eg with shrink wrap mod, but fast....
            # obj.data.transform(obj.matrix_world)
            # depsgraph.update() #fixes bad transformation baing applied to obj
            # bvh = BVHTree.FromObject(obj, depsgraph)  #? not required to get with mod: obj.evaluated_get(depsgraph)
            # obj.data.transform(obj.matrix_world.inverted())
            # return bvh

            #* better but slower - even 5-10 times (0.05 sec), wont work on non meshes (curves?)
            # get merged version of scene objects with modifiers.   - skip non mesh objs
            # NOTE: this is slow, but works for all objects (even curves)

            active_obj_mw = active_obj.matrix_world  # bring it into active obj (retopo obj) space...

            if shared_xray_props.snap_target == 'OBJECT':
                snap_obj  = shared_xray_props.snap_object
                if not shared_xray_props.snap_object or not shared_xray_props.snap_object.visible_get() or shared_xray_props.snap_object == active_obj:
                    return None

                bm = bmesh.new()   # create an empty BMesh
                # handles groups, geo nodes generated objs, mods, curves? etc.
                get_nested_obj_depsgrapth_bm(bm, depsgraph, {snap_obj.name,}, apply_modifiers=True, matrix=active_obj_mw.inverted())
                bm.normal_update() # or else we get bad normals in snapping!
                bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                bm.free()  # free and prevent further access

                # WARNING: below option would say 'empty' instance coll has no mesh... even if it has one inside inst_col
                # obj_eval = snap_obj.evaluated_get(depsgraph)
                # bm = bmesh.new()   # create an empty BMesh
                # bm.from_mesh(obj_eval.to_mesh())   # with modifiers
                # bm.transform(active_obj_mw.inverted() @ snap_obj.matrix_world)
                # bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # bm.free()  # free and prevent further access
                # obj_eval.to_mesh_clear()
                print('Xray: Updated BVHTree')
                return bvh

            else:
                # NOTE: bmesh can merge multiple src meshes!
                # bm = bmesh.new()  # create an empty BMesh - and join all meshes into int
                # bm.from_mesh( mesh1 ) # Add first mesh
                # bm.from_mesh( mesh2 ) # Add 2nd mesh
                snap_objs = []
                if shared_xray_props.snap_target == 'COLLECTION':
                    if shared_xray_props.snap_collection in context.scene.collection.children_recursive:
                        snap_objs = get_collection_objects(shared_xray_props.snap_collection, active_obj)
                else:
                    snap_objs = {o.name for o in context.scene.objects if o.visible_get() and o.display_type in ('SOLID','TEXTURED') and o != active_obj}
                if not snap_objs:
                    print('Xray: No snap objects found')
                    return None

                bm = bmesh.new()
                # for o in snap_objs: # WARNING: may be slow (O^2 :  for snap_objs: for all eval objs..)
                get_nested_obj_depsgrapth_bm(bm, depsgraph, snap_objs, apply_modifiers=True)
                bm.transform(active_obj_mw.inverted())
                bm.normal_update() # or else we get bad normals in snapping!
                bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # debug_mesh(bm)
                bm.free()  # free and prevent further access

                # NOTE: ok but empties instances would not be evaluated as groups, curves - give error 'Obj does not have geometry data'
                # bm = bmesh.new()
                # eval_objs = [] # for ev_obj.to_mesh_clear
                # for o in snap_objs:
                #     if o != active_obj and o.visible_get():
                #         obj_ev = o.evaluated_get(depsgraph)
                #         eval_objs.append(obj_ev)
                #         me = obj_ev.to_mesh()
                #         me.transform(o.matrix_world)
                #         bm.from_mesh(me) # should join all meshes into one bm
                # bm.transform(active_obj_mw.inverted())
                # bm.normal_update()
                # bvh = BVHTree.FromBMesh(bm)  # ? not required to get with mod: obj.evaluated_get(depsgraph)
                # bm.free()  # free and prevent further access
                # for obj_eval in eval_objs:
                #     obj_eval.to_mesh_clear()

                print('Xray: Updated BVHTree')
                return bvh
        else:
            return BVHTree.FromObject(snap_obj, depsgraph) #with modes
    else:
        if world_space:
            # 4 times slower than data.transform
            #bvh1 =  BVHTree.FromPolygons([obj.matrix_world @ v.co for v in obj.data.vertices], [p.vertices for p in obj.data.polygons])
            #bmesh - same time as data.transform
            snap_obj.data.transform(snap_obj.matrix_world)
            bvh = BVHTree.FromPolygons([v.co for v in snap_obj.data.vertices], [p.vertices for p in snap_obj.data.polygons])
            snap_obj.data.transform(snap_obj.matrix_world.inverted())
            print('Xray: Updated BVHTree')
            return bvh
        else:
            print('Xray: Updated BVHTree')
            return BVHTree.FromPolygons([v.co for v in snap_obj.data.vertices], [p.vertices for p in snap_obj.data.polygons])


class XRAY_OT_ObjectPicker(bpy.types.Operator):
    bl_idname = "xray.object_picker"
    bl_label = "Pick Object"
    bl_description = 'Pick Object under the cursor'
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    use_scn_prop: bpy.props.BoolProperty(name='Scene prop', description='Pick scn.snap_object prop or obj.snap_object', default=True)

    @classmethod
    def poll(cls, context):
        return True

    def invoke(self, context, event):
        self.lmb_clicked = True
        self.obj = None
        context.window.cursor_modal_set('EYEDROPPER')
        context.window_manager.modal_handler_add(self)
        self.depsgraph = context.evaluated_depsgraph_get()
        return {"RUNNING_MODAL"}

    def scn_ray_cast(self, context, event):
        ''' using scn.ray_cast '''
        region = context.region
        rv3d = context.region_data
        coord = event.mouse_region_x, event.mouse_region_y
        view_vector = view3d_utils.region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = view3d_utils.region_2d_to_origin_3d(region, rv3d, coord)
        # hit, loc, norm, idx, obj, mat = context.scene.ray_cast(context.view_layer, ray_origin, view_vector, distance=1.0e+6) #old
        hit, loc, norm, idx, obj, mat = context.scene.ray_cast(self.depsgraph, ray_origin, view_vector, distance=1.0e+6)
        return obj

    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            self.obj = self.scn_ray_cast(context, event)
            context.area.header_text_set(self.obj.name if self.obj else None)

        elif event.type == "MIDDLEMOUSE":
            return {'PASS_THROUGH'}

        elif event.type == "LEFTMOUSE":
            if self.obj and self.obj.type == 'MESH':
                if self.use_scn_prop:
                    context.scene.xray_props.snap_object = self.obj
                else:
                    context.active_object.xray_props.snap_object = self.obj
                self.report({'INFO'}, f'Picked {self.obj.name}')
            else:
                self.report({'WARNING'}, 'Pick mesh type of object')
            context.window.cursor_modal_restore()
            context.area.header_text_set(None)
            return {"FINISHED"}

        elif event.type in {"RIGHTMOUSE", "ESC"}:
            context.window.cursor_modal_restore()
            context.area.header_text_set(None)
            return {"CANCELLED"}

        return {"RUNNING_MODAL"}

